@@ -25,7 +25,7 @@
  * truncated. Else they will be disallowed.
  */
 /* #define NO_TRUNCATE */
-
+	
 /*
  * ok, we cannot use strncmp, as the name is not in our data space.
  * Thus we'll have to use ext2_match. No big problem. ext2_match also makes
@@ -33,25 +33,26 @@
  *
  * NOTE! unlike strncmp, ext2_match returns 1 for success, 0 for failure.
  */
-static int ext2_match (int len, const char * name,
+static int ext2_match (int len, const char * const name,
 		       struct ext2_dir_entry * de)
 {
-	register int same __asm__("ax");
+	unsigned char same;
 
 	if (!de || !de->inode || len > EXT2_NAME_LEN)
 		return 0;
 	/* "" means "." ---> so paths like "/usr/lib//libc.a" work */
-	if (!len && (de->name[0] == '.') && (de->name[1] == '\0'))
+	if (!len && de->name_len == 1 && (de->name[0] == '.') &&
+	   (de->name[1] == '\0'))
 		return 1;
-	if (len < EXT2_NAME_LEN && len != de->name_len)
+	if (len != de->name_len)
 		return 0;
 	__asm__("cld\n\t"
 		"repe ; cmpsb\n\t"
-		"setz %%al"
-		:"=a" (same)
-		:"0" (0), "S" ((long) name), "D" ((long) de->name), "c" (len)
+		"setz %0"
+		:"=q" (same)
+		:"S" ((long) name), "D" ((long) de->name), "c" (len)
 		:"cx", "di", "si");
-	return same;
+	return (int) same;
 }
 
 /*
@@ -62,19 +63,16 @@ static int ext2_match (int len, const char * name,
  * itself (as a parameter - res_dir). It does NOT read the inode of the
  * entry - you'll have to do that yourself if you want to.
  *
- * In the ext2 file system, this function also returns a pointer on the
- * previous directory entry because functions which remove directory
- * entries need it
  */
 static struct buffer_head * ext2_find_entry (struct inode * dir,
-					     const char * name, int namelen,
-					     struct ext2_dir_entry ** res_dir,
-					     struct ext2_dir_entry ** prev_dir)
+					     const char * const name, int namelen,
+					     struct ext2_dir_entry ** res_dir)
 {
 	long offset;
 	struct buffer_head * bh;
 	struct ext2_dir_entry * de;
 	struct super_block * sb;
+	int err;
 
 	*res_dir = NULL;
 	if (!dir)
@@ -87,23 +85,20 @@ static struct buffer_head * ext2_find_entry (struct inode * dir,
 	if (namelen > EXT2_NAME_LEN)
 		namelen = EXT2_NAME_LEN;
 #endif
-	bh = ext2_bread (dir, 0, 0);
+	bh = ext2_bread (dir, 0, 0, &err);
 	if (!bh)
 		return NULL;
-	if (prev_dir)
-		*prev_dir = NULL;
 	offset = 0;
 	de = (struct ext2_dir_entry *) bh->b_data;
 	while (offset < dir->i_size) {
-		if ((char *)de >= sb->s_blocksize + bh->b_data) {
+		if (!bh || (char *)de >= sb->s_blocksize + bh->b_data) {
 			brelse (bh);
-			bh = NULL;
-			bh = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS(sb), 0);
-			if (!bh)
+			bh = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS(sb), 0, &err);
+			if (!bh) {
+				offset += sb->s_blocksize;
 				continue;
+			}
 			de = (struct ext2_dir_entry *) bh->b_data;
-			if (prev_dir)
-				*prev_dir = NULL;
 		}
 		if (! ext2_check_dir_entry ("ext2_find_entry", dir, de, bh,
 					    offset)) {
@@ -115,8 +110,6 @@ static struct buffer_head * ext2_find_entry (struct inode * dir,
 			return bh;
 		}
 		offset += de->rec_len;
-		if (prev_dir)
-			*prev_dir = de;
 		de = (struct ext2_dir_entry *) ((char *) de + de->rec_len);
 	}
 	brelse (bh);
@@ -140,7 +133,7 @@ int ext2_lookup (struct inode * dir, const char * name, int len,
 #ifndef DONT_USE_DCACHE
 	if (!(ino = ext2_dcache_lookup (dir->i_dev, dir->i_ino, name, len))) {
 #endif
-		if (!(bh = ext2_find_entry (dir, name, len, &de, NULL))) {
+		if (!(bh = ext2_find_entry (dir, name, len, &de))) {
 			iput (dir);
 			return -ENOENT;
 		}
@@ -173,7 +166,8 @@ int ext2_lookup (struct inode * dir, const char * name, int len,
  */
 static struct buffer_head * ext2_add_entry (struct inode * dir,
 					    const char * name, int namelen,
-					    struct ext2_dir_entry ** res_dir)
+					    struct ext2_dir_entry ** res_dir,
+					    int *err)
 {
 	int i;
 	long offset;
@@ -182,6 +176,7 @@ static struct buffer_head * ext2_add_entry (struct inode * dir,
 	struct ext2_dir_entry * de, * de1;
 	struct super_block * sb;
 
+	*err = -EINVAL;
 	*res_dir = NULL;
 	if (!dir)
 		return NULL;
@@ -195,29 +190,41 @@ static struct buffer_head * ext2_add_entry (struct inode * dir,
 #endif
 	if (!namelen)
 		return NULL;
-	bh = ext2_bread (dir, 0, 0);
+	/* Is this a busy deleted directory?  Can't create new files 
+	   if so */
+	if (dir->i_size == 0)
+	{
+		*err = -ENOENT;
+		return NULL;
+	}
+	bh = ext2_bread (dir, 0, 0, err);
 	if (!bh)
 		return NULL;
 	rec_len = EXT2_DIR_REC_LEN(namelen);
 	offset = 0;
 	de = (struct ext2_dir_entry *) bh->b_data;
+	*err = -ENOSPC;
 	while (1) {
 		if ((char *)de >= sb->s_blocksize + bh->b_data) {
 			brelse (bh);
 			bh = NULL;
-			bh = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS(sb), 0);
-			if (!bh) {
+			bh = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS(sb), 1, err);
+			if (!bh)
+				return NULL;
+			if (dir->i_size <= offset) {
+				if (dir->i_size == 0) {
+					*err = -ENOENT;
+					return NULL;
+				}
 #ifdef EXT2FS_DEBUG
 				printk ("ext2_add_entry: creating next block\n");
 #endif
-				bh = ext2_bread (dir, offset >> EXT2_BLOCK_SIZE_BITS(sb), 1);
-				if (!bh)
-					return NULL;
 				de = (struct ext2_dir_entry *) bh->b_data;
 				de->inode = 0;
 				de->rec_len = sb->s_blocksize;
-				dir->i_size += sb->s_blocksize;
+				dir->i_size = offset + sb->s_blocksize;
 				dir->i_dirt = 1;
+				dir->i_ctime = CURRENT_TIME;
 			} else {
 #ifdef EXT2FS_DEBUG
 				printk ("ext2_add_entry: skipping to next block\n");
@@ -227,9 +234,17 @@ static struct buffer_head * ext2_add_entry (struct inode * dir,
 		}
 		if (! ext2_check_dir_entry ("ext2_add_entry", dir, de, bh,
 					    offset)) {
+			*err = -ENOENT;
 			brelse (bh);
 			return NULL;
 		}
+		if (de->inode) {
+			if (ext2_match (namelen, name, de)) {
+				*err = -EEXIST;
+				brelse (bh);
+				return NULL;
+			}
+		}
 		if ((!de->inode && de->rec_len >= rec_len) ||
 		    (de->rec_len >= EXT2_DIR_REC_LEN(de->name_len) + rec_len)) {
 			offset += de->rec_len;
@@ -245,9 +260,10 @@ static struct buffer_head * ext2_add_entry (struct inode * dir,
 			de->name_len = namelen;
 			for (i = 0; i < namelen ; i++)
 				de->name[i] = name [i];
-			dir->i_mtime = CURRENT_TIME;
+			dir->i_mtime = dir->i_ctime = CURRENT_TIME;
 			bh->b_dirt = 1;
 			*res_dir = de;
+			*err = 0;
 			return bh;
 		}
 		offset += de->rec_len;
@@ -261,13 +277,31 @@ static struct buffer_head * ext2_add_entry (struct inode * dir,
  * ext2_delete_entry deletes a directory entry by merging it with the
  * previous entry
  */
-static void ext2_delete_entry (struct ext2_dir_entry * dir,
-			       struct ext2_dir_entry * prev_dir)
+static int ext2_delete_entry (struct ext2_dir_entry * dir,
+			      struct buffer_head *bh)
 {
-	if (prev_dir != NULL)
-		prev_dir->rec_len += dir->rec_len;
-	else
-		dir->inode = 0;
+	struct ext2_dir_entry * de, * pde;
+	int i;
+
+	i = 0;
+	pde = NULL;
+	de = (struct ext2_dir_entry *) bh->b_data;
+	while (i < bh->b_size) {
+		if (! ext2_check_dir_entry ("ext2_delete_entry", NULL, 
+					    de, bh, i))
+			return -EIO;
+		if (de == dir)  {
+			if (pde)
+				pde->rec_len += dir->rec_len;
+			else
+				dir->inode = 0;
+			return 0;
+		}
+		i += de->rec_len;
+		pde = de;
+		de = (struct ext2_dir_entry *) ((char *) de + de->rec_len);
+	}
+	return -ENOENT;
 }
 
 int ext2_create (struct inode * dir,const char * name, int len, int mode,
@@ -276,6 +310,7 @@ int ext2_create (struct inode * dir,const char * name, int len, int mode,
 	struct inode * inode;
 	struct buffer_head * bh;
 	struct ext2_dir_entry * de;
+	int err;
 
 	*result = NULL;
 	if (!dir)
@@ -288,13 +323,13 @@ int ext2_create (struct inode * dir,const char * name, int len, int mode,
 	inode->i_op = &ext2_file_inode_operations;
 	inode->i_mode = mode;
 	inode->i_dirt = 1;
-	bh = ext2_add_entry (dir, name, len, &de);
+	bh = ext2_add_entry (dir, name, len, &de, &err);
 	if (!bh) {
 		inode->i_nlink --;
 		inode->i_dirt = 1;
 		iput (inode);
 		iput (dir);
-		return -ENOSPC;
+		return err;
 	}
 	de->inode = inode->i_ino;
 #ifndef DONT_USE_DCACHE
@@ -314,10 +349,11 @@ int ext2_mknod (struct inode * dir, const char * name, int len, int mode,
 	struct inode * inode;
 	struct buffer_head * bh;
 	struct ext2_dir_entry * de;
+	int err;
 
 	if (!dir)
 		return -ENOENT;
-	bh = ext2_find_entry (dir, name, len, &de, NULL);
+	bh = ext2_find_entry (dir, name, len, &de);
 	if (bh) {
 		brelse (bh);
 		iput (dir);
@@ -350,13 +386,13 @@ int ext2_mknod (struct inode * dir, const char * name, int len, int mode,
 		inode->i_rdev = rdev;
 	inode->i_mtime = inode->i_atime = CURRENT_TIME;
 	inode->i_dirt = 1;
-	bh = ext2_add_entry (dir, name, len, &de);
+	bh = ext2_add_entry (dir, name, len, &de, &err);
 	if (!bh) {
 		inode->i_nlink --;
 		inode->i_dirt = 1;
 		iput (inode);
 		iput (dir);
-		return -ENOSPC;
+		return err;
 	}
 	de->inode = inode->i_ino;
 #ifndef DONT_USE_DCACHE
@@ -375,13 +411,20 @@ int ext2_mkdir (struct inode * dir, const char * name, int len, int mode)
 	struct inode * inode;
 	struct buffer_head * bh, * dir_block;
 	struct ext2_dir_entry * de;
-	
-	bh = ext2_find_entry (dir, name, len, &de, NULL);
+	int err;
+
+	if (!dir)
+		return -ENOENT;
+	bh = ext2_find_entry (dir, name, len, &de);
 	if (bh) {
 		brelse (bh);
 		iput (dir);
 		return -EEXIST;
 	}
+	if (dir->i_nlink >= EXT2_LINK_MAX) {
+		iput (dir);
+		return -EMLINK;
+	}
 	inode = ext2_new_inode (dir, S_IFDIR);
 	if (!inode) {
 		iput (dir);
@@ -390,13 +433,13 @@ int ext2_mkdir (struct inode * dir, const char * name, int len, int mode)
 	inode->i_op = &ext2_dir_inode_operations;
 	inode->i_size = inode->i_sb->s_blocksize;
 	inode->i_mtime = inode->i_atime = CURRENT_TIME;
-	dir_block = ext2_bread (inode, 0, 1);
+	dir_block = ext2_bread (inode, 0, 1, &err);
 	if (!dir_block) {
 		iput (dir);
 		inode->i_nlink --;
 		inode->i_dirt = 1;
 		iput (inode);
-		return -ENOSPC;
+		return err;
 	}
 	inode->i_blocks = inode->i_sb->s_blocksize / 512;
 	de = (struct ext2_dir_entry *) dir_block->b_data;
@@ -416,12 +459,12 @@ int ext2_mkdir (struct inode * dir, const char * name, int len, int mode)
 	if (dir->i_mode & S_ISGID)
 		inode->i_mode |= S_ISGID;
 	inode->i_dirt = 1;
-	bh = ext2_add_entry (dir, name, len, &de);
+	bh = ext2_add_entry (dir, name, len, &de, &err);
 	if (!bh) {
 		iput (dir);
 		inode->i_nlink = 0;
 		iput (inode);
-		return -ENOSPC;
+		return err;
 	}
 	de->inode = inode->i_ino;
 #ifndef DONT_USE_DCACHE
@@ -446,10 +489,11 @@ static int empty_dir (struct inode * inode)
 	struct buffer_head * bh;
 	struct ext2_dir_entry * de, * de1;
 	struct super_block * sb;
+	int err;
 
 	sb = inode->i_sb;
 	if (inode->i_size < EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2) ||
-	    !(bh = ext2_bread (inode, 0, 0))) {
+	    !(bh = ext2_bread (inode, 0, 0, &err))) {
 	    	printk ("warning - bad directory (dev %04x, dir %d)\n",
 			inode->i_dev, inode->i_ino);
 		return 1;
@@ -467,7 +511,7 @@ static int empty_dir (struct inode * inode)
 	while (offset < inode->i_size ) {
 		if ((void *) de >= (void *) (bh->b_data + sb->s_blocksize)) {
 			brelse (bh);
-			bh = ext2_bread (inode, offset >> EXT2_BLOCK_SIZE_BITS(sb), 1);
+			bh = ext2_bread (inode, offset >> EXT2_BLOCK_SIZE_BITS(sb), 1, &err);
 			if (!bh) {
 				offset += sb->s_blocksize;
 				continue;
@@ -495,16 +539,26 @@ int ext2_rmdir (struct inode * dir, const char * name, int len)
 	int retval;
 	struct inode * inode;
 	struct buffer_head * bh;
-	struct ext2_dir_entry * de, * pde;
+	struct ext2_dir_entry * de;
 
+repeat:
+	if (!dir)
+		return -ENOENT;
 	inode = NULL;
-	bh = ext2_find_entry (dir, name, len, &de, &pde);
+	bh = ext2_find_entry (dir, name, len, &de);
 	retval = -ENOENT;
 	if (!bh)
 		goto end_rmdir;
 	retval = -EPERM;
 	if (!(inode = iget (dir->i_sb, de->inode)))
 		goto end_rmdir;
+	if (de->inode != inode->i_ino) {
+		iput(inode);
+		brelse(bh);
+		current->counter = 0;
+		schedule();
+		goto repeat;
+	}
 	if ((dir->i_mode & S_ISVTX) && current->euid &&
 	    inode->i_uid != current->euid)
 		goto end_rmdir;
@@ -520,29 +574,31 @@ int ext2_rmdir (struct inode * dir, const char * name, int len)
 		retval = -ENOTEMPTY;
 		goto end_rmdir;
 	}
-	if (inode->i_count > 1) {
-		retval = -EBUSY;
-		goto end_rmdir;
+	if (inode->i_count > 1 && inode->i_nlink <= 2) {
+		/* Are we deleting the last instance of a busy directory?
+		   Better clean up if so. */
+		/* Make directory empty (it will be truncated when finally
+		   dereferenced).  This also inhibits ext2_add_entry. */
+		inode->i_size = 0;
+#ifndef DONT_USE_DCACHE
+		ext2_dcache_remove(inode->i_dev, inode->i_ino, ".", 1);
+		ext2_dcache_remove(inode->i_dev, inode->i_ino, "..", 2);
+#endif
 	}
 	if (inode->i_nlink != 2)
 		printk ("empty  directory has nlink!=2 (%d)\n", inode->i_nlink);
 #ifndef DONT_USE_DCACHE
 	ext2_dcache_remove (dir->i_dev, dir->i_ino, de->name, de->name_len);
 #endif
-	ext2_delete_entry (de, pde);
-#if 0	
-	if (pde)
-		pde->rec_len += de->rec_len;
-	else
-		de->inode = 0;
-#endif
+	retval = ext2_delete_entry (de, bh);
+	if (retval)
+		goto end_rmdir;
 	bh->b_dirt = 1;
 	inode->i_nlink = 0;
 	inode->i_dirt = 1;
 	dir->i_nlink --;
 	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
 	dir->i_dirt = 1;
-	retval = 0;
 end_rmdir:
 	iput (dir);
 	iput (inode);
@@ -555,15 +611,25 @@ int ext2_unlink (struct inode * dir, const char * name, int len)
 	int retval;
 	struct inode * inode;
 	struct buffer_head * bh;
-	struct ext2_dir_entry * de, * pde;
+	struct ext2_dir_entry * de;
 
+repeat:
+	if (!dir)
+		return -ENOENT;
 	retval = -ENOENT;
 	inode = NULL;
-	bh = ext2_find_entry (dir, name, len, &de, &pde);
+	bh = ext2_find_entry (dir, name, len, &de);
 	if (!bh)
 		goto end_unlink;
 	if (!(inode = iget (dir->i_sb, de->inode)))
 		goto end_unlink;
+	if (de->inode != inode->i_ino) {
+		iput(inode);
+		brelse(bh);
+		current->counter = 0;
+		schedule();
+		goto repeat;
+	}
 	retval = -EPERM;
 	if ((dir->i_mode & S_ISVTX) && !suser() &&
 	    current->euid != inode->i_uid &&
@@ -579,13 +645,9 @@ int ext2_unlink (struct inode * dir, const char * name, int len)
 #ifndef DONT_USE_DCACHE
 	ext2_dcache_remove (dir->i_dev, dir->i_ino, de->name, de->name_len);
 #endif
-	ext2_delete_entry (de, pde);
-#if 0
-	if (pde)
-		pde->rec_len += de->rec_len;
-	else
-		de->inode = 0;
-#endif
+	retval = ext2_delete_entry (de, bh);
+	if (retval)
+		goto end_unlink;
 	bh->b_dirt = 1;
 	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
 	dir->i_dirt = 1;
@@ -607,7 +669,7 @@ int ext2_symlink (struct inode * dir, const char * name, int len,
 	struct inode * inode = NULL;
 	struct buffer_head * bh = NULL, * name_block = NULL;
 	char * link;
-	int i;
+	int i, err;
 	int l;
 	char c;
 
@@ -624,13 +686,13 @@ int ext2_symlink (struct inode * dir, const char * name, int len,
 #ifdef EXT2FS_DEBUG
 		printk ("ext2_symlink: l=%d, normal symlink\n", l);
 #endif
-		name_block = ext2_bread (inode, 0, 1);
+		name_block = ext2_bread (inode, 0, 1, &err);
 		if (!name_block) {
 			iput (dir);
 			inode->i_nlink --;
 			inode->i_dirt = 1;
 			iput (inode);
-			return -ENOSPC;
+			return err;
 		}
 		link = name_block->b_data;
 	} else {
@@ -649,7 +711,7 @@ int ext2_symlink (struct inode * dir, const char * name, int len,
 	}
 	inode->i_size = i;
 	inode->i_dirt = 1;
-	bh = ext2_find_entry (dir, name, len, &de, NULL);
+	bh = ext2_find_entry (dir, name, len, &de);
 	if (bh) {
 		inode->i_nlink --;
 		inode->i_dirt = 1;
@@ -658,13 +720,13 @@ int ext2_symlink (struct inode * dir, const char * name, int len,
 		iput (dir);
 		return -EEXIST;
 	}
-	bh = ext2_add_entry (dir, name, len, &de);
+	bh = ext2_add_entry (dir, name, len, &de, &err);
 	if (!bh) {
 		inode->i_nlink --;
 		inode->i_dirt = 1;
 		iput (inode);
 		iput (dir);
-		return -ENOSPC;
+		return err;
 	}
 	de->inode = inode->i_ino;
 #ifndef DONT_USE_DCACHE
@@ -683,29 +745,30 @@ int ext2_link (struct inode * oldinode, struct inode * dir,
 {
 	struct ext2_dir_entry * de;
 	struct buffer_head * bh;
+	int err;
 
 	if (S_ISDIR(oldinode->i_mode)) {
 		iput (oldinode);
 		iput (dir);
 		return -EPERM;
 	}
-	if (oldinode->i_nlink > 32000) {
+	if (oldinode->i_nlink > EXT2_LINK_MAX) {
 		iput (oldinode);
 		iput (dir);
 		return -EMLINK;
 	}
-	bh = ext2_find_entry (dir, name, len, &de, NULL);
+	bh = ext2_find_entry (dir, name, len, &de);
 	if (bh) {
 		brelse (bh);
 		iput (dir);
 		iput (oldinode);
 		return -EEXIST;
 	}
-	bh = ext2_add_entry (dir, name, len, &de);
+	bh = ext2_add_entry (dir, name, len, &de, &err);
 	if (!bh) {
 		iput (dir);
 		iput (oldinode);
-		return -ENOSPC;
+		return err;
 	}
 	de->inode = oldinode->i_ino;
 #ifndef DONT_USE_DCACHE
@@ -771,11 +834,13 @@ static int do_ext2_rename (struct inode * old_dir, const char * old_name,
 {
 	struct inode * old_inode, * new_inode;
 	struct buffer_head * old_bh, * new_bh, * dir_bh;
-	struct ext2_dir_entry * old_de, * new_de, * pde;
+	struct ext2_dir_entry * old_de, * new_de;
 	int retval;
 
 	goto start_up;
 try_again:
+	if (new_bh && new_de)
+		ext2_delete_entry(new_de, new_bh);
 	brelse (old_bh);
 	brelse (new_bh);
 	brelse (dir_bh);
@@ -786,18 +851,20 @@ try_again:
 start_up:
 	old_inode = new_inode = NULL;
 	old_bh = new_bh = dir_bh = NULL;
-	old_bh = ext2_find_entry (old_dir, old_name, old_len, &old_de, &pde);
+	new_de = NULL;
+	old_bh = ext2_find_entry (old_dir, old_name, old_len, &old_de);
 	retval = -ENOENT;
 	if (!old_bh)
 		goto end_rename;
 	old_inode = iget (old_dir->i_sb, old_de->inode);
 	if (!old_inode)
 		goto end_rename;
+	retval = -EPERM;
 	if ((old_dir->i_mode & S_ISVTX) && 
 	    current->euid != old_inode->i_uid &&
 	    current->euid != old_dir->i_uid && !suser())
 		goto end_rename;
-	new_bh = ext2_find_entry (new_dir, new_name, new_len, &new_de, NULL);
+	new_bh = ext2_find_entry (new_dir, new_name, new_len, &new_de);
 	if (new_bh) {
 		new_inode = iget (new_dir->i_sb, new_de->inode);
 		if (!new_inode) {
@@ -810,8 +877,18 @@ start_up:
 		goto end_rename;
 	}
 	if (new_inode && S_ISDIR(new_inode->i_mode)) {
-		retval = -EEXIST;
-		goto end_rename;
+		retval = -EISDIR;
+		if (!S_ISDIR(old_inode->i_mode))
+			goto end_rename;
+		retval = -EINVAL;
+		if (subdir (new_dir, old_inode))
+			goto end_rename;
+		retval = -ENOTEMPTY;
+		if (!empty_dir (new_inode))
+			goto end_rename;
+		retval = -EBUSY;
+		if (new_inode->i_count > 1)
+			goto end_rename;
 	}
 	retval = -EPERM;
 	if (new_inode && (new_dir->i_mode & S_ISVTX) &&
@@ -819,25 +896,24 @@ start_up:
 	    current->euid != new_dir->i_uid && !suser())
 		goto end_rename;
 	if (S_ISDIR(old_inode->i_mode)) {
-		retval = -EEXIST;
-		if (new_bh)
-			goto end_rename;
-		retval = -EACCES;
-		if (!permission (old_inode, MAY_WRITE))
+		retval = -ENOTDIR;
+		if (new_inode && !S_ISDIR(new_inode->i_mode))
 			goto end_rename;
 		retval = -EINVAL;
 		if (subdir (new_dir, old_inode))
 			goto end_rename;
-		retval = -EIO;
-		dir_bh = ext2_bread (old_inode, 0, 0);
+		dir_bh = ext2_bread (old_inode, 0, 0, &retval);
 		if (!dir_bh)
 			goto end_rename;
 		if (PARENT_INO(dir_bh->b_data) != old_dir->i_ino)
 			goto end_rename;
+		retval = -EMLINK;
+		if (!new_inode && new_dir->i_nlink >= EXT2_LINK_MAX)
+			goto end_rename;
 	}
 	if (!new_bh)
-		new_bh = ext2_add_entry (new_dir, new_name, new_len, &new_de);
-	retval = -ENOSPC;
+		new_bh = ext2_add_entry (new_dir, new_name, new_len, &new_de,
+					 &retval);
 	if (!new_bh)
 		goto end_rename;
 /* sanity checking before doing the rename - avoid races */
@@ -855,19 +931,11 @@ start_up:
 	ext2_dcache_add (new_dir->i_dev, new_dir->i_ino, new_de->name,
 			 new_de->name_len, new_de->inode);
 #endif
-	if (old_bh->b_blocknr == new_bh->b_blocknr &&
-	    ((char *) new_de) + new_de->rec_len == (char *) old_de)
-		new_de->rec_len += old_de->rec_len;
-	else
-		ext2_delete_entry (old_de, pde);
-#if 0
-	if (((char *) new_de) + new_de->rec_len == (char *) old_de)
-		new_de->rec_len += old_de->rec_len;
-	else if (pde)
-		pde->rec_len += old_de->rec_len;
-	else
-		old_de->inode = 0;
-#endif
+	retval = ext2_delete_entry (old_de, old_bh);
+	if (retval == -ENOENT)
+		goto try_again;
+	if (retval)
+		goto end_rename;
 	if (new_inode) {
 		new_inode->i_nlink --;
 		new_inode->i_dirt = 1;
