@@ -47,7 +47,7 @@ int sys_sigpending(sigset_t *set)
 /*
  * atomically swap in the new signal mask, and wait for a signal.
  */
-int sys_sigsuspend(volatile int restart, volatile unsigned long oldmask, unsigned long set)
+int sys_sigsuspend(int restart, unsigned long oldmask, unsigned long set)
 {
 	unsigned long mask;
 	struct pt_regs * regs = (struct pt_regs *) &restart;
@@ -97,15 +97,17 @@ static void check_pending(int signum)
 	}	
 }
 
-int sys_signal(int signum, long handler)
+int sys_signal(int signum, unsigned long handler)
 {
 	struct sigaction tmp;
 
 	if (signum<1 || signum>32 || signum==SIGKILL || signum==SIGSTOP)
 		return -EINVAL;
+	if (handler >= TASK_SIZE)
+		return -EFAULT;
 	tmp.sa_handler = (void (*)(int)) handler;
 	tmp.sa_mask = 0;
-	tmp.sa_flags = SA_ONESHOT | SA_NOMASK | SA_INTERRUPT;
+	tmp.sa_flags = SA_ONESHOT | SA_NOMASK;
 	tmp.sa_restorer = NULL;
 	handler = (long) current->sigaction[signum-1].sa_handler;
 	current->sigaction[signum-1] = tmp;
@@ -129,6 +131,8 @@ int sys_sigaction(int signum, const struct sigaction * action,
 			new.sa_mask |= _S(signum);
 			new.sa_mask &= _BLOCKABLE;
 		}
+		if (TASK_SIZE <= (unsigned long) new.sa_handler)
+			return -EFAULT;
 	}
 	if (oldaction) {
 		if (!verify_area(VERIFY_WRITE,oldaction, sizeof(struct sigaction)))
@@ -146,95 +150,123 @@ extern int sys_waitpid(pid_t pid,unsigned long * stat_addr, int options);
 /*
  * This sets regs->esp even though we don't actually use sigstacks yet..
  */
-int sys_sigreturn(int signr, unsigned long oldmask, unsigned long esp)
+int sys_sigreturn(unsigned long oldmask, unsigned long eip, unsigned long esp)
 {
 	struct pt_regs * regs;
 
-	regs = (struct pt_regs *) &signr;
+	regs = (struct pt_regs *) &oldmask;
 	current->blocked = oldmask & _BLOCKABLE;
+	regs->eip = eip;
 	regs->esp = esp;
 	return 0;
 }
 
 /*
- * This routine sets up the return stack for the first signal found
- * (== last delivered). It makes room for the registers we need to save,
- * but the actual saving is left until the very last moment when we
- * know whether we can restart system calls etc.
+ * Set up a signal frame... Make the stack look the way iBCS2 expects
+ * it to look.
  */
-static unsigned long * setup_first(struct pt_regs * regs,
-	int signr, unsigned long sa_handler, unsigned long oldmask)
-{
-	unsigned long * tmp_esp;
-
-	regs->esp -= 18*4;
-	tmp_esp = (unsigned long *) regs->esp;
-	verify_area(VERIFY_WRITE,tmp_esp,18*4);
-/* set up the "normal" stack seen by the signal handler */
-	put_fs_long(regs->esp+15*4,tmp_esp);	/* points to the stack.. */
-	put_fs_long(signr,tmp_esp+1);		/* parameter to handler and sigreturn */
-	put_fs_long((unsigned long) (tmp_esp+5),tmp_esp+2);
-	put_fs_long(oldmask,tmp_esp+3);		/* second .. */
-	put_fs_long(__NR_sigreturn,tmp_esp+4);	/* sigreturn number.. */
-/* save this frame so that we later can fill in the saved registers */
-	return tmp_esp+5;
-}
-
-/*
- * This sets up the stack for any stacked signals other than the
- * first one: no need to restore registers etc, as that is done
- * by the very last signal handler return code..
- */
-static void setup_other(unsigned long eip, struct pt_regs * regs, int signr,
+static void setup_frame(unsigned long ** fp, unsigned long eip,
+	struct pt_regs * regs, int signr,
 	unsigned long sa_handler, unsigned long oldmask)
 {
-	unsigned long * tmp_esp;
+	unsigned long * frame;
 
-	regs->esp -= 9*4;
-	tmp_esp = (unsigned long *) regs->esp;
-	verify_area(VERIFY_WRITE,tmp_esp,9*4);
-/* set up the "normal" stack seen by the signal handler */
-	put_fs_long(regs->esp+6*4,tmp_esp);	/* points to the stack.. */
-	put_fs_long(signr,tmp_esp+1);		/* parameter to handler and sigreturn */
-	put_fs_long((unsigned long) (tmp_esp+5),tmp_esp+2);
-	put_fs_long(oldmask,tmp_esp+3);		/* second .. */
-	put_fs_long(__NR_sigreturn,tmp_esp+4);	/* sigreturn number.. */
-	put_fs_long(eip,tmp_esp+5);		/* return address */
+#define __CODE ((unsigned long)(frame+24))
+#define CODE(x) ((void *) ((x)+__CODE))
+	frame = *fp - 32;
+	verify_area(VERIFY_WRITE,frame,32*4);
+/* set up the "normal" stack seen by the signal handler (iBCS2) */
+	put_fs_long(__CODE,frame);
+	put_fs_long(signr, frame+1);
+	put_fs_long(regs->gs, frame+2);
+	put_fs_long(regs->fs, frame+3);
+	put_fs_long(regs->es, frame+4);
+	put_fs_long(regs->ds, frame+5);
+	put_fs_long(regs->edi, frame+6);
+	put_fs_long(regs->esi, frame+7);
+	put_fs_long(regs->ebp, frame+8);
+	put_fs_long(regs->esp, frame+9);
+	put_fs_long(regs->ebx, frame+10);
+	put_fs_long(regs->edx, frame+11);
+	put_fs_long(regs->ecx, frame+12);
+	put_fs_long(regs->eax, frame+13);
+	put_fs_long(0, frame+14);		/* trapno */
+	put_fs_long(0, frame+15);		/* err */
+	put_fs_long(regs->eip, frame+16);
+	put_fs_long(regs->cs, frame+17);
+	put_fs_long(regs->eflags, frame+18);
+	put_fs_long(regs->esp, frame+19);
+	put_fs_long(regs->ss, frame+20);
+	put_fs_long(0,frame+21);		/* 387 state pointer */
+/* linux extended stack - easier to handle.. */
+	put_fs_long(regs->eflags, frame+22);
+	put_fs_long(eip, frame+23);
 /* set up the return code... */
-	put_fs_long(0x58595a5b,tmp_esp+6);	/* pop bx,dx,cx,ax */
-	put_fs_long(0x909080cd,tmp_esp+7);	/* int $0x80 + nop + nop */
-	put_fs_long(0x000cc290,tmp_esp+8);	/* nop + "ret 12" */
+	put_fs_long(0x0000b858, CODE(0));	/* popl %eax ; movl $,%eax */
+	put_fs_long(0x00bb0000, CODE(4));	/* movl $,%ebx */
+	put_fs_long(0xcd000000, CODE(8));	/* int $0x80 */
+	put_fs_long(0x0fa90f80, CODE(12));	/* pop %gs ; pop %fs */
+	put_fs_long(0x611f07a1, CODE(16));	/* pop %es ; pop %ds ; popad */
+	put_fs_long(0x20c48390, CODE(20));	/* nop ; addl $32,%esp */
+	put_fs_long(0x0020c29d, CODE(24));	/* popfl ; ret $32 */
+	put_fs_long(__NR_ssetmask, CODE(2));
+	put_fs_long(oldmask, CODE(7));
+	*fp = frame;
+#undef __CODE
+#undef CODE
 }
 
 /*
  * Note that 'init' is a special process: it doesn't get signals it doesn't
  * want to handle. Thus you cannot kill init even with a SIGKILL even by
  * mistake.
+ *
+ * Note that we go through the signals twice: once to check the signals that
+ * the kernel can handle, and then we build all the user-level signal handling
+ * stack-frames in one go after that.
  */
 int do_signal(unsigned long oldmask, struct pt_regs * regs)
 {
+	unsigned long mask = ~current->blocked;
+	unsigned long handler_signal = 0;
 	unsigned long *frame = NULL;
 	unsigned long eip = 0;
 	unsigned long signr;
 	unsigned long sa_handler;
 	struct sigaction * sa;
 
-	while ((signr = current->signal & ~current->blocked)) {
+	while ((signr = current->signal & mask)) {
 		__asm__("bsf %2,%1\n\t"
 			"btrl %1,%0"
 			:"=m" (current->signal),"=r" (signr)
 			:"1" (signr));
 		sa = current->sigaction + signr;
 		signr++;
-		sa_handler = (unsigned long) sa->sa_handler;
-		if (sa_handler==1) {
-/* check for SIGCHLD: it's special */
-			if (signr == SIGCHLD)
-				while (sys_waitpid(-1,NULL,WNOHANG) > 0)
-					/* nothing */;
+		if ((current->flags & PF_PTRACED) && signr != SIGKILL) {
+			current->exit_code = signr;
+			current->state = TASK_STOPPED;
+			notify_parent(current);
+			schedule();
+			if (!(signr = current->exit_code))
+				continue;
+			current->exit_code = 0;
+			if (signr == SIGSTOP)
+				continue;
+			if (_S(signr) & current->blocked) {
+				current->signal |= _S(signr);
+				continue;
+			}
+			sa = current->sigaction + signr - 1;
+		}
+		if (sa->sa_handler == SIG_IGN) {
+			if (signr != SIGCHLD)
+				continue;
+			/* check for SIGCHLD: it's special */
+			while (sys_waitpid(-1,NULL,WNOHANG) > 0)
+				/* nothing */;
 			continue;
 		}
-		if (!sa_handler) {
+		if (sa->sa_handler == SIG_DFL) {
 			if (current->pid == 1)
 				continue;
 			switch (signr) {
@@ -246,7 +278,7 @@ int do_signal(unsigned long oldmask, struct pt_regs * regs)
 				current->exit_code = signr;
 				if (!(current->p_pptr->sigaction[SIGCHLD-1].sa_flags & 
 						SA_NOCLDSTOP))
-					send_sig(SIGCHLD, current->p_pptr, 1);
+					notify_parent(current);
 				schedule();
 				continue;
 
@@ -265,20 +297,11 @@ int do_signal(unsigned long oldmask, struct pt_regs * regs)
 		 */
 		if (regs->orig_eax >= 0) {
 			if (regs->eax == -ERESTARTNOHAND ||
-			   (regs->eax == -ERESTARTSYS && (sa->sa_flags & SA_INTERRUPT)))
+			   (regs->eax == -ERESTARTSYS && !(sa->sa_flags & SA_RESTART)))
 				regs->eax = -EINTR;
 		}
-		if (sa->sa_flags & SA_ONESHOT)
-			sa->sa_handler = NULL;
-/* force a supervisor-mode page-in of the signal handler to reduce races */
-		__asm__("testb $0,%%fs:%0"::"m" (*(char *) sa_handler));
-		if (!frame) {
-			frame = setup_first(regs,signr,sa_handler,oldmask);
-		} else
-			setup_other(eip,regs,signr,sa_handler,oldmask);
-		eip = sa_handler;
-		current->blocked |= sa->sa_mask;
-		oldmask |= sa->sa_mask;
+		handler_signal |= 1 << (signr-1);
+		mask &= ~sa->sa_mask;
 	}
 	if (regs->orig_eax >= 0 &&
 	    (regs->eax == -ERESTARTNOHAND ||
@@ -287,23 +310,28 @@ int do_signal(unsigned long oldmask, struct pt_regs * regs)
 		regs->eax = regs->orig_eax;
 		regs->eip -= 2;
 	}
-	if (!frame)				/* no handlers installed - return 0 */
+	if (!handler_signal)		/* no handler will be called - return 0 */
 		return 0;
-/* save registers if one or more handlers are called.. */
-	put_fs_long(regs->edi,frame);		/* suitable order for "popad" */
-	put_fs_long(regs->esi,frame+1);
-	put_fs_long(regs->ebp,frame+2);		/* using 'frame++' instead of the 'frame+x' */
-	put_fs_long(regs->esp,frame+3);		/* form used now results in atrocious code */
-	put_fs_long(regs->ebx,frame+4);		/* due to gcc not being very good at optimizing */
-	put_fs_long(regs->edx,frame+5);		/* things with inline-assembly/functions.. */
-	put_fs_long(regs->ecx,frame+6);
-	put_fs_long(regs->eax,frame+7);
-	put_fs_long(regs->eflags,frame+8);	/* flags */
-	put_fs_long(regs->eip,frame+9);		/* original return address */
-/* set up the return code... */
-	put_fs_long(0x58595a5b,frame+10);	/* pop bx,dx,cx,ax */
-	put_fs_long(0x906180cd,frame+11);	/* int $0x80 + popad + nop */
-	put_fs_long(0x000cc29d,frame+12);	/* popfl + "ret 12" */
+	eip = regs->eip;
+	frame = (unsigned long *) regs->esp;
+	signr = 1;
+	sa = current->sigaction;
+	for (mask = 1 ; mask ; sa++,signr++,mask += mask) {
+		if (mask > handler_signal)
+			break;
+		if (!(mask & handler_signal))
+			continue;
+		sa_handler = (unsigned long) sa->sa_handler;
+		if (sa->sa_flags & SA_ONESHOT)
+			sa->sa_handler = NULL;
+/* force a supervisor-mode page-in of the signal handler to reduce races */
+		__asm__("testb $0,%%fs:%0"::"m" (*(char *) sa_handler));
+		setup_frame(&frame,eip,regs,signr,sa_handler,oldmask);
+		eip = sa_handler;
+		current->blocked |= sa->sa_mask;
+		oldmask |= sa->sa_mask;
+	}
+	regs->esp = (unsigned long) frame;
 	regs->eip = eip;			/* "return" to the first handler */
 	return 1;
 }
